# -*- coding: utf-8 -*-
"""file module"""

__author__  = "Adrien DELLE CAVE"
__license__ = """
    Copyright (C) 2017-2018  fjord-technologies

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA..
"""

import logging
import os
import resource
import threading

from dwho.adapters.redis import DWhoAdapterRedis
from dwho.classes.modules import DWhoModuleBase, MODULES

from hashlib import sha1
from httpdis.ext.httpdis_json import HttpResponseJson, HttpReqErrJson
from sonicprobe.libs import xys


LOG             = logging.getLogger('xtrargon.modules.file')

BUFFER_SIZE     = 1048576
RLIMIT_NOFILE   = 65536
SHA1_SZERO  =    'da39a3ee5e6b4b0d3255bfef95601890afd80709'
_CACHE          = {}


class XtrargonFileHashThread(threading.Thread):
    def __init__(self, hash_check, adapter_redis, xkey, fstat, filepath):
        threading.Thread.__init__(self)

        self.hash_check    = hash_check
        self.adapter_redis = adapter_redis
        self.xkey          = xkey
        self.fstat         = fstat
        self.filepath      = filepath
        self.xhash         = ''

    def _mk_xsum(self):
        return '|'.join((str(self.fstat.st_size),
                         str(self.fstat.st_mtime),
                         self.xhash or ''))

    def _get_file_hash(self):
        (f, xhash) = (None, None)

        try:
            if self.fstat.st_size == 0:
                return SHA1_SZERO

            with open(self.filepath, 'rb') as f:
                xhash = sha1()

                while True:
                    buf = f.read(BUFFER_SIZE)
                    if not buf:
                        break
                    xhash.update(buf)

            LOG.debug("finished to calculate hash. (filepath: %r)", self.filepath)

            return "%s" % xhash.hexdigest()
        except (IOError, OSError), e:
            LOG.error("unable to read file: %r. (error: %r)", self.filepath, e)
        finally:
            if f:
                f.close()
            f, xhash = (None, None)

    def run(self):
        if self.hash_check:
            self.xhash = self._get_file_hash()
        else:
            self.xhash = ''
        xsum = self._mk_xsum()
        self.adapter_redis.set_key(self.xkey, xsum)
        return xsum


class XtrargonFileModule(DWhoModuleBase):
    MODULE_NAME = 'file'

    def safe_init(self, options):
        self.hash_check     = bool(int(self.config['general'].get('hash_check', False)))
        self.is_full        = False
        self.server_id      = self.config['general']['server_id']

        self.adapter_redis  = DWhoAdapterRedis(self.config, prefix = 'file_check')
        resource.setrlimit(resource.RLIMIT_NOFILE, (RLIMIT_NOFILE, RLIMIT_NOFILE))

        if not self.adapter_redis.servers:
            raise SystemError("missing redis server")

    def _mk_his_key(self, his, full = False):
        if not full:
            key = "hincr"
        else:
            key = "hfull"

        return "%s:%s:%s" % (self.server_id, key, his)

    def _mk_lsn_key(self, lsn, full = False):
        if not full:
            key = "lincr"
        else:
            key = "lfull"

        return "%s:%s:%s" % (self.server_id, key, lsn)

    def _fetch_his_full(self, his):
        key = self._mk_his_key(his)

        if key in _CACHE:
            return _CACHE[key]

        fhis = self.adapter_redis.get_key(key)

        if fhis:
            r = self._fetch_his_full(fhis)
        else:
            r = self.adapter_redis.get_key(self._mk_his_key(his, True))

        if r:
            _CACHE[key] = r

        return r

    def _set_his(self, options):
        if not options.get('history'):
            return

        if options.get('history-incr'):
            ret = self._fetch_his_full(options['history-incr'])
            self.adapter_redis.set_key(self._mk_his_key(options['history']),
                                       ret)
            return ret

        self.adapter_redis.set_key(self._mk_his_key(options['history'], True),
                                   options['history'])
        return options['history']

    def _fetch_lsn_full(self, his):
        key = self._mk_lsn_key(his)

        if key in _CACHE:
            return _CACHE[key]

        fhis = self.adapter_redis.get_key(key)

        if fhis:
            r = self._fetch_lsn_full(fhis)
        else:
            r = self.adapter_redis.get_key(self._mk_lsn_key(his, True))

        if r:
            _CACHE[key] = r

        return r

    def _set_lsn(self, options):
        if options['from-lsn'] != '0':
            ret = self._fetch_lsn_full(options['from-lsn'])
            self.adapter_redis.set_key(self._mk_lsn_key(options['from-lsn']),
                                       ret)
            return ret

        self.adapter_redis.set_key(self._mk_lsn_key(options['from-lsn'], True),
                                   options['from-lsn'])
        return options['from-lsn']


    CHECK_SCHEMA = xys.load("""
    history*-: !!str
    history-incr*-: !!str
    from-lsn!-: !~~digit
    filepath!-: !!str
    """)

    def check(self, request):
        options      = request.query_params()
        if not xys.validate(options, self.CHECK_SCHEMA):
            raise HttpReqErrJson(400, "invalid request")

        self.is_full = int(options['from-lsn']) is 0

        try:
            hiskey  = self._set_his(options) or self._set_lsn(options)
            if not hiskey:
                raise HttpReqErrJson(500, "unable to find or to set history key. (file: %r)"
                                          % options['filepath'])

            LOG.info("history key found: %s", hiskey)

            xkey    = "%s:%s:%s" % (self.server_id, hiskey, options['filepath'])
            fstat   = os.stat(options['filepath'])

            t       = XtrargonFileHashThread(self.hash_check, self.adapter_redis, xkey, fstat, options['filepath'])
            t.setDaemon(True)

            if self.is_full:
                LOG.debug("full detected so no file check. (key: %r)", xkey)
                t.start()
                return HttpResponseJson(code = 205)

            xsum    = self.adapter_redis.get_key(xkey)
            if not xsum:
                LOG.debug("unable to find file in redis. (key: %r)", xkey)
                t.start()
                return HttpResponseJson(code = 205)

            xlist   = xsum.split('|')

            if xlist[0] != str(fstat.st_size):
                LOG.debug("filesize not equal. (key: %r, previous: %r, current: %r)",
                          xkey,
                          xlist[0],
                          str(fstat.st_size))
                t.start()
                return HttpResponseJson(code = 205)

            if xlist[1] != str(fstat.st_mtime):
                LOG.debug("filetime not equal. (key: %r, previous: %r, current: %r)",
                          xkey,
                          xlist[1],
                          str(fstat.st_mtime))
                t.start()
                return HttpResponseJson(code = 205)

            t.setDaemon(False)
            nxsum = t.run()

            if not nxsum:
                raise HttpReqErrJson(404, "unable to find file or calculate hash. (file: %r)"
                                          % options['filepath'])

            if xsum != nxsum:
                LOG.debug("hash not equal. (key: %r, previous: %r, current: %r)",
                          xkey,
                          xsum,
                          nxsum)
                return HttpResponseJson(code = 205)

            LOG.debug("hash equal. (key: %r, previous: %r, current: %r)",
                      xkey,
                      xsum,
                      nxsum)
            return HttpResponseJson(code = 304)
        except HttpReqErrJson:
            raise
        except Exception, e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))


if __name__ != "__main__":
    def _start():
        MODULES.register(XtrargonFileModule())
    _start()
